@@ -3,12 +3,15 @@
 module Generics.SOP.TH
   ( deriveGeneric
   , deriveGenericOnly
+  , deriveGenericSubst
+  , deriveGenericOnlySubst
   , deriveGenericFunctions
   , deriveMetadataValue
   , deriveMetadataType
   ) where
 
 import Control.Monad (replicateM)
+import Data.List (foldl')
 import Data.Maybe (fromMaybe)
 import Data.Proxy
 import Language.Haskell.TH
@@ -51,7 +54,7 @@ import Generics.SOP.Universe
 -- >
 -- >   to (SOP    (Z (I x :* Nil)))         = Leaf x
 -- >   to (SOP (S (Z (I l :* I r :* Nil)))) = Node l r
--- >   to _ = error "unreachable" -- to avoid GHC warnings
+-- >   to (SOP (S (S x)))                   = x `seq` error "inaccessible"
 -- >
 -- > instance HasDatatypeInfo Tree where
 -- >   type DatatypeInfoOf Tree =
@@ -66,17 +69,33 @@ import Generics.SOP.Universe
 -- datatypes with unboxed fields.
 --
 deriveGeneric :: Name -> Q [Dec]
-deriveGeneric n = do
-  dec <- reifyDec n
-  ds1 <- withDataDec dec deriveGenericForDataDec
-  ds2 <- withDataDec dec deriveMetadataForDataDec
-  return (ds1 ++ ds2)
+deriveGeneric n =
+  deriveGenericSubst n varT
 
 -- | Like 'deriveGeneric', but omit the 'HasDatatypeInfo' instance.
 deriveGenericOnly :: Name -> Q [Dec]
-deriveGenericOnly n = do
+deriveGenericOnly n =
+  deriveGenericOnlySubst n varT
+
+-- | Variant of 'deriveGeneric' that allows to restrict the type parameters.
+--
+-- Experimental function, exposed primarily for benchmarking.
+--
+deriveGenericSubst :: Name -> (Name -> Q Type) -> Q [Dec]
+deriveGenericSubst n f = do
   dec <- reifyDec n
-  withDataDec dec deriveGenericForDataDec
+  ds1 <- withDataDec dec (deriveGenericForDataDec  f)
+  ds2 <- withDataDec dec (deriveMetadataForDataDec f)
+  return (ds1 ++ ds2)
+
+-- | Variant of 'deriveGenericOnly' that allows to restrict the type parameters.
+--
+-- Experimental function, exposed primarily for benchmarking.
+--
+deriveGenericOnlySubst :: Name -> (Name -> Q Type) -> Q [Dec]
+deriveGenericOnlySubst n f = do
+  dec <- reifyDec n
+  withDataDec dec (deriveGenericForDataDec f)
 
 -- | Like 'deriveGenericOnly', but don't derive class instance, only functions.
 --
@@ -95,7 +114,7 @@ deriveGenericOnly n = do
 -- > toTree :: SOP I TreeCode -> Tree
 -- > toTree (SOP    (Z (I x :* Nil)))         = Leaf x
 -- > toTree (SOP (S (Z (I l :* I r :* Nil)))) = Node l r
--- > toTree _ = error "unreachable" -- to avoid GHC warnings
+-- > toTree (SOP (S (S x)))                   = x `seq` error "inaccessible"
 --
 -- @since 0.2
 --
@@ -106,9 +125,9 @@ deriveGenericFunctions n codeName fromName toName = do
   let toName'   = mkName toName
   dec <- reifyDec n
   withDataDec dec $ \_isNewtype _cxt name bndrs cons _derivs -> do
-    let codeType = codeFor cons                        -- '[ '[Int], '[Tree, Tree] ]
-    let origType = appTyVars name bndrs                -- Tree
-    let repType  = [t| SOP I $(appTyVars codeName' bndrs) |] -- SOP I TreeCode
+    let codeType = codeFor varT cons                     -- '[ '[Int], '[Tree, Tree] ]
+    let origType = appTyVars varT name bndrs             -- Tree
+    let repType  = [t| SOP I $(appTyVars varT codeName' bndrs) |] -- SOP I TreeCode
     sequence
       [ tySynD codeName' bndrs codeType                 -- type TreeCode = '[ '[Int], '[Tree, Tree] ]
       , sigD fromName' [t| $origType -> $repType |]     -- fromTree :: Tree -> SOP I TreeCode
@@ -166,23 +185,29 @@ deriveMetadataType n datatypeInfoName = do
     sequence
       [ tySynD datatypeInfoName' [] (metadataType' isNewtype name cons) ]
 
-deriveGenericForDataDec :: Bool -> Cxt -> Name -> [TyVarBndr] -> [Con] -> Derivings -> Q [Dec]
-deriveGenericForDataDec _isNewtype _cxt name bndrs cons _derivs = do
-  let typ = appTyVars name bndrs
-#if MIN_VERSION_template_haskell(2,9,0)
-  let codeSyn = tySynInstD ''Code $ tySynEqn [typ] (codeFor cons)
-#else
-  let codeSyn = tySynInstD ''Code [typ] (codeFor cons)
-#endif
+deriveGenericForDataDec ::
+  (Name -> Q Type) -> Bool -> Cxt -> Name -> [TyVarBndr] -> [Con] -> Derivings -> Q [Dec]
+deriveGenericForDataDec f _isNewtype _cxt name bndrs cons _derivs = do
+  let typ = appTyVars f name bndrs
+  deriveGenericForDataType f typ cons
+
+deriveGenericForDataType :: (Name -> Q Type) -> Q Type -> [Con] -> Q [Dec]
+deriveGenericForDataType f typ cons = do
+  let codeSyn = tySynInstD ''Code $ tySynEqn [typ] (codeFor f cons)
   inst <- instanceD
             (cxt [])
             [t| Generic $typ |]
             [codeSyn, embedding 'from cons, projection 'to cons]
   return [inst]
 
-deriveMetadataForDataDec :: Bool -> Cxt -> Name -> [TyVarBndr] -> [Con] -> Derivings -> Q [Dec]
-deriveMetadataForDataDec isNewtype _cxt name bndrs cons _derivs = do
-  let typ = appTyVars name bndrs
+deriveMetadataForDataDec ::
+  (Name -> Q Type) -> Bool -> Cxt -> Name -> [TyVarBndr] -> [Con] -> Derivings -> Q [Dec]
+deriveMetadataForDataDec f isNewtype _cxt name bndrs cons _derivs = do
+  let typ = appTyVars f name bndrs
+  deriveMetadataForDataType isNewtype name typ cons
+
+deriveMetadataForDataType :: Bool -> Name -> Q Type -> [Con] -> Q [Dec]
+deriveMetadataForDataType isNewtype name typ cons = do
   md   <- instanceD (cxt [])
             [t| HasDatatypeInfo $typ |]
             [ metadataType typ isNewtype name cons
@@ -195,17 +220,16 @@ deriveMetadataForDataDec isNewtype _cxt name bndrs cons _derivs = do
             -- [metadata isNewtype name cons]
   return [md]
 
-
 {-------------------------------------------------------------------------------
   Computing the code for a data type
 -------------------------------------------------------------------------------}
 
-codeFor :: [Con] -> Q Type
-codeFor = promotedTypeList . map go
+codeFor :: (Name -> Q Type) -> [Con] -> Q Type
+codeFor f = promotedTypeList . map go
   where
     go :: Con -> Q Type
     go c = do (_, ts) <- conInfo c
-              promotedTypeList ts
+              promotedTypeListSubst f ts
 
 {-------------------------------------------------------------------------------
   Computing the embedding/projection pair
@@ -233,31 +257,44 @@ embedding fromName = funD fromName . go' (\e -> [| Z $e |])
              []
 
 projection :: Name -> [Con] -> Q Dec
-projection toName = funD toName . go' (\p -> conP 'Z [p])
+projection toName = funD toName . go'
   where
-    go' :: (Q Pat -> Q Pat) -> [Con] -> [Q Clause]
-    go' _ [] = (:[]) $ do
+    go' :: [Con] -> [Q Clause]
+    go' [] = (:[]) $ do
       x <- newName "x"
       clause [varP x] (normalB (caseE (varE x) [])) []
-    go' br cs = go br cs
+    go' cs = go id cs
 
     go :: (Q Pat -> Q Pat) -> [Con] -> [Q Clause]
-    go _ [] = [unreachable]
+    go br [] = [mkUnreachableClause br]
     go br (c:cs) = mkClause br c : go (\p -> conP 'S [br p]) cs
 
+    -- Generates a final clause of the form:
+    --
+    --   to (S (... (S x))) = x `seq` error "inaccessible"
+    --
+    -- An equivalent way of achieving this would be:
+    --
+    --   to (S (... (S x))) = case x of {}
+    --
+    -- This, however, would require clients to enable the EmptyCase extension
+    -- in their own code, which is something which we have not previously
+    -- required. Therefore, we do not generate this code at the moment.
+    mkUnreachableClause :: (Q Pat -> Q Pat) -> Q Clause
+    mkUnreachableClause br = do
+      var <- newName "x"
+      clause [conP 'SOP [br (varP var)]]
+             (normalB [| $(varE var) `seq` error "inaccessible" |])
+             []
+
     mkClause :: (Q Pat -> Q Pat) -> Con -> Q Clause
     mkClause br c = do
       (n, ts) <- conInfo c
       vars    <- replicateM (length ts) (newName "x")
-      clause [conP 'SOP [br . npP . map (\v -> conP 'I [varP v]) $ vars]]
+      clause [conP 'SOP [br . conP 'Z . (:[]) . npP . map (\v -> conP 'I [varP v]) $ vars]]
              (normalB . appsE $ conE n : map varE vars)
              []
 
-unreachable :: Q Clause
-unreachable = clause [wildP]
-                     (normalB [| error "unreachable" |])
-                     []
-
 {-------------------------------------------------------------------------------
   Compute metadata
 -------------------------------------------------------------------------------}
@@ -287,24 +324,13 @@ metadata' isNewtype typeName cs = md
                                                $(npE (map mdField ts))
                              |]
     mdCon (InfixC _ n _)  = do
-#if MIN_VERSION_template_haskell(2,11,0)
       fixity <- reifyFixity n
       case fromMaybe defaultFixity fixity of
         Fixity f a ->
-#else
-      i <- reify n
-      case i of
-        DataConI _ _ _ (Fixity f a) ->
-#endif
                             [| SOP.Infix       $(stringE (nameBase n)) $(mdAssociativity a) f |]
-#if !MIN_VERSION_template_haskell(2,11,0)
-        _                -> fail "Strange infix operator"
-#endif
     mdCon (ForallC _ _ _) = fail "Existentials not supported"
-#if MIN_VERSION_template_haskell(2,11,0)
     mdCon (GadtC _ _ _)    = fail "GADTs not supported"
     mdCon (RecGadtC _ _ _) = fail "GADTs not supported"
-#endif
 
     mdField :: VarStrictType -> Q Exp
     mdField (n, _, _) = [| SOP.FieldInfo $(stringE (nameBase n)) |]
@@ -335,24 +361,13 @@ metadataType' isNewtype typeName cs = md
                                                    $(promotedTypeList (map mdField ts))
                               |]
     mdCon (InfixC _ n _)  = do
-#if MIN_VERSION_template_haskell(2,11,0)
       fixity <- reifyFixity n
       case fromMaybe defaultFixity fixity of
         Fixity f a ->
-#else
-      i <- reify n
-      case i of
-        DataConI _ _ _ (Fixity f a) ->
-#endif
                             [t| 'SOP.T.Infix       $(stringT (nameBase n)) $(mdAssociativity a) $(natT f) |]
-#if !MIN_VERSION_template_haskell(2,11,0)
-        _                -> fail "Strange infix operator"
-#endif
     mdCon (ForallC _ _ _) = fail "Existentials not supported"
-#if MIN_VERSION_template_haskell(2,11,0)
     mdCon (GadtC _ _ _)    = fail "GADTs not supported"
     mdCon (RecGadtC _ _ _) = fail "GADTs not supported"
-#endif
 
     mdField :: VarStrictType -> Q Type
     mdField (n, _, _) = [t| 'SOP.T.FieldInfo $(stringT (nameBase n)) |]
@@ -394,10 +409,8 @@ conInfo (NormalC n ts) = return (n, map (return . (\(_, t)    -> t)) ts)
 conInfo (RecC    n ts) = return (n, map (return . (\(_, _, t) -> t)) ts)
 conInfo (InfixC (_, t) n (_, t')) = return (n, map return [t, t'])
 conInfo (ForallC _ _ _) = fail "Existentials not supported"
-#if MIN_VERSION_template_haskell(2,11,0)
 conInfo (GadtC _ _ _)    = fail "GADTs not supported"
 conInfo (RecGadtC _ _ _) = fail "GADTs not supported"
-#endif
 
 stringT :: String -> Q Type
 stringT = litT . strTyLit
@@ -409,13 +422,36 @@ promotedTypeList :: [Q Type] -> Q Type
 promotedTypeList []     = promotedNilT
 promotedTypeList (t:ts) = [t| $promotedConsT $t $(promotedTypeList ts) |]
 
-appTyVars :: Name -> [TyVarBndr] -> Q Type
-appTyVars n = go (conT n)
+promotedTypeListSubst :: (Name -> Q Type) -> [Q Type] -> Q Type
+promotedTypeListSubst _ []     = promotedNilT
+promotedTypeListSubst f (t:ts) = [t| $promotedConsT $(t >>= substType f) $(promotedTypeListSubst f ts) |]
+
+appsT :: Name -> [Q Type] -> Q Type
+appsT n = foldl' appT (conT n)
+
+bndrToName :: TyVarBndr -> Name
+bndrToName (PlainTV  v  ) = v
+bndrToName (KindedTV v _) = v
+
+appTyVars :: (Name -> Q Type) -> Name -> [TyVarBndr] -> Q Type
+appTyVars f n bndrs =
+  appsT n (map (f . bndrToName) bndrs)
+
+substType :: (Name -> Q Type) -> Type -> Q Type
+substType f = go
   where
-    go :: Q Type -> [TyVarBndr] -> Q Type
-    go t []                  = t
-    go t (PlainTV  v   : vs) = go [t| $t $(varT v) |] vs
-    go t (KindedTV v _ : vs) = go [t| $t $(varT v) |] vs
+    go (VarT n)     = f n
+    go (AppT t1 t2) = AppT <$> go t1 <*> go t2
+    go ListT        = return ListT
+    go (ConT n)     = return (ConT n)
+    go ArrowT       = return ArrowT
+    go (TupleT i)   = return (TupleT i)
+    go t            = return t -- error (show t)
+      -- TODO: This is incorrect, but we only need substitution to work
+      -- in simple cases for now. The reason is that substitution is normally
+      -- the identity, except if we use TH derivation for the tagged datatypes
+      -- in the benchmarking suite. So we can fall back on identity in all
+      -- but the cases we need for the benchmarking suite.
 
 reifyDec :: Name -> Q Dec
 reifyDec name =
@@ -424,20 +460,13 @@ reifyDec name =
                   _          -> fail "Info must be type declaration type."
 
 withDataDec :: Dec -> (Bool -> Cxt -> Name -> [TyVarBndr] -> [Con] -> Derivings -> Q a) -> Q a
-#if MIN_VERSION_template_haskell(2,11,0)
 withDataDec (DataD    ctxt name bndrs _ cons derivs) f = f False ctxt name bndrs cons  derivs
 withDataDec (NewtypeD ctxt name bndrs _ con  derivs) f = f True  ctxt name bndrs [con] derivs
-#else
-withDataDec (DataD    ctxt name bndrs cons derivs) f = f False ctxt name bndrs cons  derivs
-withDataDec (NewtypeD ctxt name bndrs con  derivs) f = f True  ctxt name bndrs [con] derivs
-#endif
 withDataDec _ _ = fail "Can only derive labels for datatypes and newtypes."
 
 -- | Utility type synonym to cover changes in the TH code
 #if MIN_VERSION_template_haskell(2,12,0)
 type Derivings = [DerivClause]
-#elif MIN_VERSION_template_haskell(2,11,0)
-type Derivings = Cxt
 #else
-type Derivings = [Name]
+type Derivings = Cxt
 #endif
